/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.objectweb.petals.tools.rmi.common.serializable;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessagingException;
import javax.security.auth.Subject;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.stream.InputStreamForker;

/**
 * The normalized message implementation.
 * @author Nicolas Salatge - EBM WebSourcing
 */
public class NormalizedMessage implements javax.jbi.messaging.NormalizedMessage, Serializable {

    /**
     * COMPRESSION constants.
     */
    public static final String COMPRESSION = "org.objectweb.petals.transport.compress";

    /**
     * the serialVersionUID.
     */
    private static final long serialVersionUID = 1L;

    /**
     * Attachments to the main message: id + content.
     */
    private transient Map<String, DataHandler> attachments;

    /**
     * The XML "payload" message.
     */
    private transient Source content;

    /**
     * Metadata associated with the message content. Map: property + value.
     */
    private Map<String, Object> properties;

    /**
     * JAAS Security Subject.
     */
    private Subject subject;

    /**
     * The input forked stream to preserve availability of the content
     */
    private transient InputStreamForker forkedInputStreamContent;

    /**
     * This default constructor creates a new empty
     * <code>NormalizedMessage</code>.
     * 
     * @throws MessagingException
     *      impossible to create a fault
     */
    public NormalizedMessage() {
        this.content = null;
        this.properties = new HashMap<String, Object>();
        this.attachments = new HashMap<String, DataHandler>();
    }

    /*
     * ---------------------------------------------------------------------- N
     * N o r m a l i z e d M e s s a g e M e t h o d s
     * ----------------------------------------------------------------------
     */

    /**
     * @inheritDoc
     * @param id
     *      the id
     * @param attachment
     *      the attachement
     * @throws MessagingException
     *      impossible to add an attachement
     */
    public void addAttachment(final String id, final DataHandler attachment)
            throws MessagingException {
        this.attachments.put(id, attachment);
    }

    /**
     * @inheritDoc
     * @param id
     *      thye id
     * @return the data handler
     */
    public DataHandler getAttachment(final String id) {
        return (DataHandler) this.attachments.get(id);
    }

    /**
     * @inheritDoc
     * @return the set of keys
     */
    public Set<?> getAttachmentNames() {
        return this.attachments.keySet();
    }

    /**
     * @inheritDoc
     * @return the source
     */
    public Source getContent() {
        if (this.content instanceof StreamSource) {
            this.forkedInputStreamContent = new InputStreamForker(((StreamSource) this.content)
                    .getInputStream());

            try {
                this.content = new StreamSource(this.forkedInputStreamContent.fork());
            } catch (final IOException e) {
                throw new UncheckedException(e);
            }
        }

        return this.content;
    }

    /**
     * @inheritDoc
     * @param name
     *      the name
     * @return the object property
     */
    public Object getProperty(final String name) {
        return this.properties.get(name);
    }

    /**
     * @inheritDoc
     * @return the set of properties
     */
    public Set<?> getPropertyNames() {
        return this.properties.keySet();
    }

    /**
     * @inheritDoc
     * @return the subject
     */
    public Subject getSecuritySubject() {
        return this.subject;
    }

    /**
     * @inheritDoc
     * @param id
     *      the id
     * @throws MessagingException
     *      impossible to remove the attachment
     */
    public void removeAttachment(final String id) throws MessagingException {
        if (this.attachments.remove(id) == null) {
            throw new MessagingException(id + " attachment does not exist.");
        }
    }

    /**
     * @inheritDoc
     * @param content
     *      the content
     * @throws MessagingException
     *      impossible to set the content
     */
    public void setContent(final Source content) throws MessagingException {
        this.content = content;
    }

    /**
     * @inheritDoc
     * @param name
     *      the name
     * @param value
     *      the value
     */
    public void setProperty(final String name, final Object value) {
        if (value == null) {
            // case to remove properties not defined (null)
            this.properties.remove(name);

        } else {
            // define this property name with a new value
            this.properties.put(name, value);
        }
    }

    /**
     * @inheritDoc
     * @param securitySubject
     *      the security subject
     */
    public void setSecuritySubject(final Subject securitySubject) {
        this.subject = securitySubject;
    }

    /**
     * @see NormalizedMessageImpl#readObject(ObjectInputStream).
     * @param s
     *      the object input stream
     * @throws IOException
     *      impossible to read object delegate
     */
    protected void readObjectDelegate(final ObjectInputStream s) throws IOException {
        try {
            s.defaultReadObject();

                this.attachments = MessageExchangeSerializer.instance().deserializeAttachments(s);
                String compression = (String) this.getProperty(NormalizedMessage.COMPRESSION);
                this.content = MessageExchangeSerializer.instance().deserializeContent(s,
                        Boolean.valueOf(compression));
        } catch (ClassNotFoundException e) {
            throw new IOException(e.getClass() + ":" + e.getMessage());
        }
    }

    /**
     * Serialization process : <br/> - Serialization of all serializable fields
     * of the NormalizedMessage <br/> - Serialization of all attachement <br/> -
     * Finally, serialization of the content. Content must always be serialized
     * at the end of the serialization process. Because in the deserialization
     * process, a new content will be created with all ending bytes of the
     * Stream.
     * 
     * @param s
     *      the object output stream
     * 
     * @see NormalizedMessageImpl#writeObject(ObjectOutputStream)
     * 
     * @throws IOException
     *      impossible to write object delegate      
     */
    protected void writeObjectDelegate(final ObjectOutputStream s) throws IOException {
        s.defaultWriteObject();

        MessageExchangeSerializer.instance().serializeAttachments(this.attachments, s);

        String compression = (String) this.getProperty(NormalizedMessage.COMPRESSION);

        if (this.forkedInputStreamContent != null) {
            final InputStream is = this.forkedInputStreamContent.fork();
                this.content = new StreamSource(is);
        }

        MessageExchangeSerializer.instance().serializeContent(this.content, s,
                Boolean.valueOf(compression));

        if (this.forkedInputStreamContent != null) {
            this.content = new StreamSource(this.forkedInputStreamContent.fork());
        }
    }
    
    /**
     * Used to deserialize Content object.
     * 
     * @param s
     *      the object output stream
     * @throws IOException
     *      impossible to read object
     */
    private void readObject(final ObjectInputStream s) throws IOException {
        this.readObjectDelegate(s);
    }

    /**
     * Used to serialize Content object.
     * 
     * @param s
     *      the object output stream
     * @throws IOException
     *      impossible to write object
     */
    private void writeObject(final ObjectOutputStream s) throws IOException {
        this.writeObjectDelegate(s);
    }
}
